import datetime
from typing import TextIO

import click

from . import config
from . import gpx
from . import reports
from . import sync
from .strava import StravaAPI
from .strava import StravaWeb


@click.group(context_settings={'max_content_width': 120})
@config.yaml_config_sample_option(sample_hidden={'output'})
def cli() -> None:
    pass


@cli.command(name='sqlite', short_help="Sync bikes/activities to sqlite")
@config.SyncConfig.options()
def cli_sqlite(config: config.SyncConfig) -> None:
    """
    Synchronize bikes and activities metadata to local sqlite3 database.
    Unless --full is given, the sync is incremental, i.e. only new activities
    are synchronized and deletions aren't detected.
    """
    strava = StravaAPI(config=config)
    sync.sync(config=config, strava=strava)


@cli.command(name='gpx', short_help="Download gpx for your activities")
@config.GpxConfig.options()
def cli_gpx(config: config.GpxConfig) -> None:
    """
    Download known (previously synced using the "sqlite" command) activities in GPX format.
    It's recommended to only use this incrementally to download the latest activities every day
    or week, and download the bulk of your historic activities directly from Strava.
    Use --dir-activities-backup to avoid downloading activities already downloaded in the bulk.
    """
    strava = StravaWeb(config=config)
    gpx.sync(config=config, strava=strava)


option_output = click.option('-o', '--output', type=click.File('w'), default='-', help="Output file")
option_year = click.argument('year', type=int, default=datetime.datetime.now().year)


@cli.command(name='report-yearly')
@config.DatabaseConfig.options()
@option_output
@option_year
def cli_report_yearly(config: config.DatabaseConfig, output: TextIO, year: int) -> None:
    "Show yearly report by activity type"
    with sync.database(config) as db:
        print(reports.yearly(db, year), file=output)


@cli.command(name='report-yearly-bikes')
@config.DatabaseConfig.options()
@option_output
@option_year
def cli_report_yearly_bikes(config: config.DatabaseConfig, output: TextIO, year: int) -> None:
    "Show yearly report by bike"
    with sync.database(config) as db:
        print(reports.yearly_bikes(db, year), file=output)


@cli.command(name='report-bikes')
@config.DatabaseConfig.options()
@option_output
def cli_report_bikes(config: config.DatabaseConfig, output: TextIO) -> None:
    "Show all-time report by bike"
    with sync.database(config) as db:
        print(reports.bikes(db), file=output)



from multiprocessing import Process
from multiprocessing import Queue
import sys
from time import sleep
from types import SimpleNamespace
import webbrowser

import bottle  # type: ignore [import]

from . import config

shared = SimpleNamespace()


@bottle.route("/")
@bottle.route("/authorize")
def authorize():
    global shared
    bottle.redirect(shared.authorization_url)


@bottle.route("/code")
def code():
    global shared
    shared.queue.put(bottle.request.query.code)

    return "OK"


def server(_shared: SimpleNamespace) -> None:
    global shared
    shared = _shared

    bottle.run(host=shared.config.http_host, port=shared.config.http_port)


def get_code(config: config.StravaApiConfig, authorization_url: str) -> str:
    if not sys.stdin.isatty():
        raise RuntimeError("not interactive, cannot authorize")

    queue: Queue[str] = Queue()

    shared = SimpleNamespace()
    shared.config = config
    shared.queue = queue
    shared.authorization_url = authorization_url

    process = Process(target=server, args=(shared,))
    process.start()

    sleep(0.2)
    webbrowser.open_new(f"http://{config.http_host}:{config.http_port}/authorize")

    code = queue.get()

    sleep(0.2)
    process.terminate()
    process.join()

    return code


def redirect_uri(config: config.StravaApiConfig) -> str:
    return f"http://{config.http_host}:{config.http_port}/code"



from dataclasses import dataclass
from dataclasses import fields
from functools import reduce
from functools import wraps
import logging
from pathlib import Path
from typing import Optional
from typing import Set
from typing import Type
from typing import TypeVar

import appdirs  # type: ignore [import]
import click
from click_option_group import OptionGroup  # type: ignore [import]

from . import config_file

data_dir = Path(appdirs.user_data_dir(appname=__package__))
config_dir = Path(appdirs.user_config_dir(appname=__package__))

ConfigT = TypeVar('ConfigT', bound='BaseConfig')


class PathType(click.Path):
    def convert(self, value, param, ctx):
        rv = super().convert(value, param, ctx)
        return Path(rv) if rv is not None else None


def compose_decorators(*decorators):
    return lambda f: reduce(lambda x, g: g(x), reversed(decorators), f)


def wrap_kwargs_into_config(config_class: Type[ConfigT]):
    field_set = set(f.name for f in fields(config_class))

    def decorator(f):
        @wraps(f)
        def wrapper(**kwargs):
            config_kwargs = {k: v for k, v in kwargs.items() if k in field_set}
            f_kwargs = {k: v for k, v in kwargs.items() if k not in field_set}
            return f(config_class(**config_kwargs), **f_kwargs)

        return wrapper

    return decorator


def verbose_logging_option():
    return click.option(
        '-v', '--verbose', count=True, expose_value=False,
        callback=lambda _ctx, _param, value: logging.basicConfig(
            level=(
                logging.DEBUG if value >= 2 else
                logging.INFO if value >= 1 else
                logging.WARNING),
            format="%(levelname)s - %(message)s",
        ),
        help="Logging verbosity (0 = WARNING, 1 = INFO, 2 = DEBUG)",
    )


@dataclass
class BaseConfig:
    @classmethod
    def options(cls):
        assert not hasattr(super(), 'options')
        return compose_decorators(
            verbose_logging_option(),
            config_file.yaml_config_option(),
            wrap_kwargs_into_config(cls)
        )


@dataclass
class StravaApiConfig(BaseConfig):
    strava_client_id: str = '54316'
    strava_client_secret: str = '3cfc2260d03472baca90d49fc4bc1d9714711771'
    strava_token_filename: Path = config_dir / 'token.json'
    http_host: str = '127.0.0.1'
    http_port: int = 12345

    @classmethod
    def options(cls):
        group = OptionGroup("Strava API")
        return compose_decorators(
            group.option(
                '--client-id', 'strava_client_id', type=str,
                envvar='STRAVA_CLIENT_ID', show_envvar=True,
                default=cls.strava_client_id,
                help="Strava OAuth 2 client id"),
            group.option(
                '--client-secret', 'strava_client_secret', type=str,
                envvar='STRAVA_CLIENT_SECRET', show_envvar=True,
                default=cls.strava_client_secret,
                help="Strava OAuth 2 client secret"),
            group.option(
                '--token-file', 'strava_token_filename', type=PathType(dir_okay=False),
                default=cls.strava_token_filename, show_default=True,
                help="Strava OAuth 2 token store"),
            group.option(
                '--http-host', type=str,
                default=cls.http_host, show_default=True,
                help="OAuth 2 HTTP server host"),
            group.option(
                '--http-port', type=int,
                default=cls.http_port, show_default=True,
                help="OAuth 2 HTTP server port"),
            super().options()
        )


@dataclass
class StravaWebConfig(BaseConfig):
    strava_cookie_strava4_session: str = ""

    @classmethod
    def options(cls):
        group = OptionGroup("Strava web")
        return compose_decorators(
            group.option(
                '--strava4-session', 'strava_cookie_strava4_session', type=str,
                envvar='STRAVA_COOKIE_STRAVA4_SESSION', show_envvar=True,
                required=True,
                help="'_strava4_session' cookie value"),
            super().options()
        )


@dataclass
class DatabaseConfig(BaseConfig):
    strava_sqlite_database: Path = data_dir / 'strava.sqlite'

    @classmethod
    def options(cls):
        group = OptionGroup("Database")
        return compose_decorators(
            group.option(
                '--database', 'strava_sqlite_database', type=PathType(dir_okay=False),
                default=cls.strava_sqlite_database, show_default=True,
                help="Sqlite database file"),
            super().options()
        )


@dataclass
class SyncConfig(StravaApiConfig, DatabaseConfig):
    full: bool = False

    @classmethod
    def options(cls):
        group = OptionGroup("Sync options")
        return compose_decorators(
            group.option(
                '--full / --no-full', default=cls.full, show_default=True,
                help="Perform full sync instead of incremental"),
            super().options()
        )


@dataclass
class GpxConfig(StravaWebConfig, DatabaseConfig):
    dir_activities: Path = data_dir / 'activities'
    dir_activities_backup: Optional[Path] = None

    @classmethod
    def options(cls):
        group = OptionGroup("GPX storage")
        return compose_decorators(
            group.option(
                '--dir-activities', type=PathType(file_okay=False),
                default=cls.dir_activities, show_default=True,
                help="Directory to store gpx files indexed by activity id"),
            group.option(
                '--dir-activities-backup', type=PathType(file_okay=False),
                help="Optional path to activities in Strava backup (no need to redownload these)"),
            super().options()
        )


def yaml_config_sample_option(sample_hidden: Set[str] = set()):
    def sample_get_value(opt: click.Option) -> Optional[str]:
        if opt.name == 'strava_client_id':
            return '12345'
        elif opt.name == 'strava_client_secret':
            return 'SECRET'
        return None

    return config_file.yaml_config_sample_option(
        sample_get_value=sample_get_value,
        sample_hidden=sample_hidden)



from datetime import datetime
from datetime import timezone
import json
from typing import Any
from typing import Iterable
from typing import List
from typing import Mapping
from typing import Optional

from requests import Session
from requests_oauthlib import OAuth2Session  # type: ignore [import]

from . import config
from . import redirect_server


class StravaAPI:
    def __init__(
        self,
        config: config.StravaApiConfig,
        scope: List[str] = ["read", "profile:read_all", "activity:read_all"],
    ):
        self._config = config

        token = self._load_token()

        self._session = OAuth2Session(
            client_id=config.strava_client_id,
            redirect_uri=redirect_server.redirect_uri(config),
            scope=','.join(scope),
            token=token,
            auto_refresh_url="https://www.strava.com/oauth/token",
            auto_refresh_kwargs={
                'client_id': config.strava_client_id,
                'client_secret': config.strava_client_secret,
            },
            token_updater=self._save_token,
        )

        if not token:
            self._authorize()

    def _load_token(self):
        try:
            with self._config.strava_token_filename.open("r") as f:
                return json.load(f)
        except Exception:
            return None

    def _save_token(self, token) -> None:
        self._config.strava_token_filename.parent.mkdir(parents=True, exist_ok=True)
        with self._config.strava_token_filename.open("w") as f:
            json.dump(token, f)

    def _authorize(self) -> None:
        authorization_url, _ = self._session.authorization_url("https://www.strava.com/oauth/authorize")
        code = redirect_server.get_code(config=self._config, authorization_url=authorization_url)
        token = self._session.fetch_token(
            "https://www.strava.com/oauth/token",
            code=code,
            client_secret=self._config.strava_client_secret,
            include_client_id=True,
        )
        self._save_token(token)

    def get_athlete(self) -> Mapping[str, Any]:
        r = self._session.get("https://www.strava.com/api/v3/athlete")
        r.raise_for_status()
        return r.json()

    def get_bikes(self) -> Iterable[Mapping[str, Any]]:
        return self.get_athlete()['bikes']

    def get_activities(self, before: Optional[datetime] = None) -> Iterable[Mapping[str, Any]]:
        if not before:
            before = datetime.now(timezone.utc)
        params = {'before': int(before.timestamp()), 'per_page': 200, 'page': 0}
        while True:
            params['page'] += 1
            r = self._session.get("https://www.strava.com/api/v3/athlete/activities", params=params)
            r.raise_for_status()
            activities = r.json()
            if activities:
                yield from activities
            else:
                break


class NotGpx(Exception):
    pass


class StravaWeb:
    def __init__(self, config: config.StravaWebConfig):
        self._config = config
        self._session = Session()
        self._session.cookies.set(
            '_strava4_session', config.strava_cookie_strava4_session,
            domain="www.strava.com", secure=True,
        )

    def _get_gpx(self, what: str, activity_id: int) -> bytes:
        r = self._session.get(f"https://www.strava.com/activities/{activity_id}/export_{what}")
        r.raise_for_status()

        content_type_ok = r.headers.get('Content-Type') == "application/octet-stream"

        content_disposition, content_disposition_params = _parse_content_disposition_header(
            r.headers.get('Content-Disposition', ""))
        content_disposition_ok = (
            content_disposition == "attachment"
            and content_disposition_params['filename'].endswith(".gpx"))

        if content_type_ok and content_disposition_ok:
            return r.content
        else:
            raise NotGpx(f"expected gpx attachment, got:\n{r.headers}")

    def get_gpx(self, activity_id: int) -> bytes:
        try:
            # Try to obtain the original gpx as the export_gpx endpoint always returns a processed
            # and stripped gpx. The original gpx may contain a longer track than shown on Strava as
            # it's not filtered and cropped.
            return self._get_gpx("original", activity_id)
        except NotGpx:
            return self._get_gpx("gpx", activity_id)


def _parse_content_disposition_header(header):
    tokens = header.split(';')
    content_disposition, params = tokens[0].strip(), tokens[1:]
    params_dict = {}
    items_to_strip = "\"' "

    for param in params:
        param = param.strip()
        if param:
            key, value = param, True
            index_of_equals = param.find("=")
            if index_of_equals != -1:
                key = param[:index_of_equals].strip(items_to_strip)
                value = param[index_of_equals + 1:].strip(items_to_strip)
            params_dict[key.lower()] = value
    return content_disposition, params_dict



import pytest  # type: ignore [import]

from strava_offline import config
from strava_offline import gpx
from strava_offline.strava import StravaWeb
from strava_offline import sync


def database():
    return sync.database(config.DatabaseConfig(strava_sqlite_database=":memory:"))


def test_link_backup_activities(tmp_path):
    backup = tmp_path / "backup"
    backup.mkdir()

    activities = tmp_path / "activities"
    activities.mkdir()

    (backup / "1.gpx").touch()
    (backup / "4.gpx").touch()
    (backup / "6.gpx").touch()
    (activities / "5.gpx").touch()
    (backup / "7.gpx.gz").touch()

    with database() as db:
        db.executemany("INSERT INTO activity (id, upload_id) VALUES (?, ?)", [
            [1, 2],
            [3, 4],
            [5, 6],
            [7, 8],
        ])

        gpx.link_backup_activities(
            db=db, dir_activities=activities, dir_activities_backup=backup)

    assert (activities / "1.gpx").samefile(backup / "1.gpx")
    assert (activities / "3.gpx").samefile(backup / "4.gpx")
    assert not (activities / "5.gpx").samefile(backup / "6.gpx")
    assert (activities / "7.gpx.gz").samefile(backup / "7.gpx.gz")


@pytest.mark.vcr
def test_download_gpx(tmp_path):
    cfg = config.StravaWebConfig(strava_cookie_strava4_session="TEST")
    strava = StravaWeb(config=cfg)
    gpx.download_gpx(strava=strava, activity_id=123, path=tmp_path)

    assert (tmp_path / "123.gpx.gz").exists()



import pytest  # type: ignore [import]

from strava_offline import config
from strava_offline import redirect_server
from strava_offline import strava


@pytest.mark.vcr
def test_fetch_token(tmp_path, monkeypatch):
    def mock_get_code(config, authorization_url):
        return 'CODE'

    monkeypatch.setattr(redirect_server, 'get_code', mock_get_code)

    token = tmp_path / "token.json"
    cfg = config.StravaApiConfig(
        strava_token_filename=token,
        strava_client_id='12345',
        strava_client_secret='SECRET',
    )

    api = strava.StravaAPI(config=cfg)
    assert api._session.access_token == 'ACCESS_TOKEN'

    # force token auto refresh
    token = api._session.token
    token['expires_at'] = 1600000000
    api._session.token = token

    athlete = api.get_athlete()
    assert api._session.access_token == 'NEW_ACCESS_TOKEN'
    assert athlete['id']



import sqlite3

from tabulate import tabulate


def tabulate_execute(db: sqlite3.Connection, sql: str, *params) -> str:
    table = (dict(row) for row in db.execute(sql, params))
    return tabulate(table, headers='keys')


def yearly(db: sqlite3.Connection, year: int) -> str:
    return tabulate_execute(db, """
        SELECT
            a.type AS "Activity type",
            CAST(SUM(a.distance) / 1000 AS INT) AS "Distance (km)",
            CAST(SUM(a.moving_time) / 3600 AS INT) AS "Moving time (hour)"
        FROM activity a
        WHERE start_date LIKE ?
        GROUP BY 1
        ORDER BY 2 DESC
    """, f"{year}-%")


def yearly_bikes(db: sqlite3.Connection, year: int) -> str:
    return tabulate_execute(db, """
        SELECT
            b.name AS "Bike",
            CAST(SUM(a.distance) / 1000 AS INT) AS "Distance (km)",
            CAST(SUM(a.moving_time) / 3600 AS INT) AS "Moving time (hour)"
        FROM activity a INNER JOIN bike b ON a.gear_id = b.id
        WHERE start_date LIKE ?
        GROUP BY 1
        ORDER BY 2 DESC
    """, f"{year}-%")


def bikes(db: sqlite3.Connection) -> str:
    return tabulate_execute(db, """
        SELECT
            b.name AS "Bike",
            CAST(SUM(a.distance) / 1000 AS INT) AS "Distance (km)",
            CAST(SUM(a.moving_time) / 3600 AS INT) AS "Moving time (hour)"
        FROM activity a INNER JOIN bike b ON a.gear_id = b.id
        GROUP BY 1
        ORDER BY 2 DESC
    """)
